# -*- coding: utf-8 -*-

"""Various water models."""

import math

import seamm_util


class Water(object):
    """Generic water model"""

    def __init__(self, r0, theta0, qO=None, qH=None, lj_a=None, lj_b=None):
        """Initialize the basic water model

        Parameters
        ----------
        r0 : float
            The O-H bond-length
        theta0 : float
            The H-O-H angle, in degrees
        qO : float, optional
            The point charge on the oxygen, defaults to None
        qH : float, optional
            The point charge on the hydrogens, defaults to None
        """
        self.r0 = r0
        self.theta0 = theta0
        self.qO = qO
        self.qH = qH
        self.lj_a = lj_a
        self.lj_b = lj_b

    @staticmethod
    def create_model(model_name):
        """Static method to simplify constructing models for string names."""
        name = model_name.lower()
        if name == "spc":
            return SPC()
        if name == "spce" or name == "spc/e":
            return SPC_E()
        if name == "tip3p":
            return TIP3P()

        raise ValueError(
            "Water model '{}' not implemented or recognized.".format(model_name)
        )

    @staticmethod
    def find_waters(system):
        """Find the water molecules in the given system.

        At the moment, the atoms are 0-based, but the bonds are given using
        1-based atom numbers!

        Parameters
        ----------
        system : System object

        Returns
        -------

        """
        elements = system["atoms"]["elements"]
        n_atoms = len(elements)

        # List the bonds, ordering the two atoms.
        bonds = []
        for i, j, order in system["bonds"]:
            i -= 1
            j -= 1
            if i < j:
                bonds.append((i, j))
            else:
                bonds.append((j, i))

        # atoms bonded to each atom i
        bonded_to = []
        for i in range(n_atoms):
            bonded_to.append([])
        for i, j in bonds:
            bonded_to[i].append(j)
            bonded_to[j].append(i)

        n_bonds = []
        for i in range(n_atoms):
            bonded_to[i].sort()
            n_bonds.append(len(bonded_to[i]))

        # Now find water molecules by looking for an O attached to two H's
        waters = []
        for i in range(n_atoms):
            if elements[i] == "O" and n_bonds[i] == 2:
                h1, h2 = bonded_to[i]
                if (
                    elements[h1] == "H"
                    and n_bonds[h1] == 1
                    and elements[h2] == "H"
                    and n_bonds[h2] == 1
                ):
                    if h1 < h1:
                        waters.append((i, h1, h2))
                    else:
                        waters.append((i, h2, h1))
        return waters

    @property
    def mass(self):
        mass_oxygen = seamm_util.element_data["O"]["atomic weight"]
        mass_hydrogen = seamm_util.element_data["H"]["atomic weight"]

        return mass_oxygen + 2 * mass_hydrogen

    def coordinates(self):
        """A standard set of coordinates in Angstrom

        The oxygen is at the origin, with the hydrogens in the x-z plane.
        Z is the two-fold axis of symmetry.

        Returns
        -------
        Tuple of three tuples with (x, y z)
        """
        # H locations are ±x, 0, z
        x = self.r0 * math.sin(math.radians(self.theta0 / 2))
        z = self.r0 * math.cos(math.radians(self.theta0 / 2))

        return ((0.0, 0.0, 0.0), (x, 0.0, z), (-x, 0.0, z))

    def pdb(self):
        """The contents of a PDB file for this model.

        Returns
        -------
        pdb : str
            The contents of a PDB file for this model.
        """
        # H locations are ±x, 0, z
        x = self.r0 * math.sin(math.radians(self.theta0 / 2))
        z = self.r0 * math.cos(math.radians(self.theta0 / 2))

        string = (
            "COMPND      Water\n"
            "HETATM    1  O           1       0.000   0.000   0.000  1.00  0.00\n"  # noqa: E501
            "HETATM    2  H           1       {x:.3f}   0.000   {z:.3f}  1.00  0.00\n"  # noqa: E501
            "HETATM    3  H           1      -{x:.3f}   0.000   {z:.3f}  1.00  0.00\n"  # noqa: E501
            "TER       4              1 \n"
            "END"
        )

        return string.format(x=x, z=z)

    def system(self):
        """Return the model as a SEAMM system.

        Returns
        -------
        system : SEAMM system
        """
        system = {
            "periodicity": 0,
            "atoms": {
                "names": ["O", "H1", "H2"],
                "elements": ["O", "H", "H"],
                "coordinates": list(self.coordinates()),
                "charges": {"*": [self.qO, self.qH, self.qH]},
                "formal charges": [0, 0, 0],
                "atom_types": {"*": self.atom_types()},
            },
            "bonds": [(1, 2, "single"), (1, 3, "single")],
            "units": {"coordinates": "angstrom"},
        }  # yapf: disable
        return system


class SPC(Water):
    def __init__(self):
        super().__init__(1.0, 109.47, -0.82, 0.41)

    def atom_types(self):
        """The atom types for this model.

        Returns
        -------
        types : str[3]
            3-vector of atom types
        """
        return ["o_spc", "h_spc", "h_spc"]


class SPC_E(Water):
    def __init__(self):
        super().__init__(1.0, 109.47, -0.8476, 0.4238)

    def atom_types(self):
        """The atom types for this model.

        Returns
        -------
        types : str[3]
            3-vector of atom types
        """
        return ["o_spc/e", "h_spc/e", "h_spc/e"]

    def reference(self):
        ris = """
            TY  - JOUR
            T1  - The missing term in effective pair potentials
            AU  - Berendsen, H. J. C.
            AU  - Grigera, J. R.
            AU  - Straatsma, T. P.
            Y1  - 1987/11/01
            PY  - 1987
            DA  - 1987/11/01
            N1  - doi: 10.1021/j100308a038
            DO  - 10.1021/j100308a038
            T2  - The Journal of Physical Chemistry
            JF  - The Journal of Physical Chemistry
            JO  - J. Phys. Chem.
            SP  - 6269
            EP  - 6271
            VL  - 91
            IS  - 24
            PB  - American Chemical Society
            SN  - 0022-3654
            M3  - doi: 10.1021/j100308a038
            UR  - https://doi.org/10.1021/j100308a038
            ER  -
        """
        return ris


class TIP3P(Water):
    def __init__(self):
        super().__init__(0.9572, 104.52, -0.834, 0.417)

    def atom_types(self):
        """The atom types for this model.

        Returns
        -------
        types : str[3]
            3-vector of atom types
        """
        return ["o_tip3p", "h_tip3p", "h_tip3p"]



# -*- coding: utf-8 -*-

"""Context manager for extending file reading to handle include's"""

import bz2
import collections
import gzip
import logging
from pathlib import Path

logger = logging.getLogger(__name__)

# logger.setLevel(logging.DEBUG)


def default_uri_handler(path):
    """Do nothing! Just return the path"""
    return Path(path)


def splitext(path):
    """
    Get the extension of a file, ignoring .gz or .bz2 on the end

    Parameters
    ----------
    path : str or pathlib.Path
     The name or path of the file
    """
    if not isinstance(path, Path):
        path = Path(path)

    ext = path.suffix
    if ext in (".gz", ".bz2"):
        ext = path.with_suffix("").suffix
    return ext


class Open(object):
    def __init__(
        self,
        path,
        mode="r",
        logger=logger,
        include="#include",
        history=10,
        uri_handler=None,
    ):
        """Open a file, automatically handling 'include'

        Parameters
        ----------
        path : str or pathlib.Path
            The path to the file
        mode : str (optional)
            The mode to open the file. Defaults to read-only.
        logger : logging.Logger (optional)
            The logging object to use
        include : str (optional)
            The keyword that triggers an include. Defaults to "#include"
        history : int (optional)
            Length of history to keep
        uri_handler : function (optional)
            A method to handle any URIs, like 'local:'. Defaults to None.
        """
        if not isinstance(path, Path):
            path = Path(path)
        self.mode = mode
        self.logger = logger
        self.include = include
        self._history = history
        self._depth = -1
        if uri_handler is None:
            self._uri_handler = default_uri_handler
        else:
            self._uri_handler = uri_handler

        self._deque = collections.deque(maxlen=history)

        self._total_lines = 0
        self._linenos = []
        self._files = []
        self._paths = [self._uri_handler(path).expanduser().resolve()]
        self._visited = [("", self.path)]
        self._fds = []
        self._cwd = Path.cwd()

    def __enter__(self):
        """Handle the enter event for the context manager by opening the file"""
        self.logger.debug("in __enter__")
        self._linenos.append(0)
        self._fds.append(self._open(self.path))

        self.logger.debug(f"   opened {self.path}")
        return self

    def __exit__(self, *args, **kwargs):
        """Close any open files as we exit the context"""
        self.logger.debug("in __exit__")
        while len(self._fds) > 0:
            fd = self._fds.pop()
            self._close(fd, *args, **kwargs)
        self.logger.debug("   closed all files")

    def __next__(self):
        """Iterator to get the next line"""
        self.logger.log(0, "__next__")

        if self._depth >= 0:
            line = self._deque[self._depth]
            self._depth -= 1
            return line

        line = self._next()
        words = line.split()
        if self.include is not None and len(words) > 0 and words[0] == self.include:
            if len(words) > 2:
                filename, tmp = words[1:3]
                missing_ok = tmp.lower() == "missing_ok"
            else:
                filename = words[1]
                missing_ok = False
            self.logger.debug("   opening include file {}".format(filename))
            self._linenos.append(0)
            try:
                path = self.path.parent / self._uri_handler(filename)
            except FileNotFoundError:
                self.logger.debug("   URI handler did not find the file")
                if not missing_ok:
                    raise
            else:
                # only use a file once
                if path not in self.visited:
                    # Check that the file exists in case the URI handler doesn't
                    if path.exists():
                        self._paths.append(path.expanduser().resolve())
                        self._fds.append(self._open(self.path))
                        self.logger.debug("   opened it")
                        self.visited.append((filename, path))
                    elif missing_ok:
                        pass
                    else:
                        raise FileNotFoundError(str(path))
            line = self.__next__()
        self.logger.log(0, line)

        self._deque.appendleft(line)
        return line

    def __iter__(self):
        """Need to be an iterator"""
        return self

    def __getattr__(self, attr):
        """Pass any attribute requests to the actual file handle"""
        self.logger.debug("attr = '{}'".format(attr))
        return getattr(self._fds[-1], attr)

    @property
    def depth(self):
        """The depth of the current line stack"""
        return self._depth

    @property
    def path(self):
        """The path of the current working file"""
        if len(self.paths) > 0:
            return self.paths[-1]
        else:
            return None

    @property
    def paths(self):
        """The paths of all the opened files."""
        return self._paths

    @property
    def lineno(self):
        """THe line number in the current file."""
        if len(self._linenos) > 0:
            return self._linenos[-1]
        else:
            return 0

    @property
    def total_lines(self):
        return self._total_lines

    @property
    def visited(self):
        """The set of files used"""
        return self._visited

    def push(self, n=1):
        """Push the n last lines back onto the device (virtually)"""
        if self._depth + n > len(self._deque):
            raise RuntimeError("Exceeded the currently available history")
        self._depth += n

    def stack(self):
        """Provide the traceback of the included files"""
        result = []
        for path, lineno in zip(reversed(self.paths), reversed(self._linenos)):
            result.append(f"{path}:{lineno}")
        return result

    def _close(self, fd, *args, **kwargs):
        """Helper routine to close a file handle"""
        exit = getattr(fd, "__exit__", None)
        if exit is not None:
            self.logger.debug("   closing fd using its __exit__ method")
            return exit(*args, **kwargs)
        else:
            exit = getattr(fd, "close", None)
            if exit is not None:
                self.logger.debug("   closing fd using its close method")
                exit()

    def _next(self):
        """Helper routine to get the next line, handling EOF and errors"""
        try:
            # Get the next line from the current file
            line = self._fds[-1].__next__()
        except StopIteration:
            # Hit EOF, so go back to previous file, if any
            fd = self._fds.pop()
            self._close(fd)
            self.paths.pop()
            n = self._linenos.pop()
            self.logger.info("   read {} lines from fd".format(n))
            if len(self._fds) <= 0:
                # At the end of the first file, so all done
                raise StopIteration()
            # and get the next line from the previous file
            return self._next()

        self._linenos[-1] += 1
        self._total_lines += 1

        return line

    def _open(self, path):
        """Open 'filename' using gzip or bzip if needed

        Parameters
        ----------
        path : pathlib.Path
            The path to the file
        """
        if not isinstance(path, Path):
            raise RuntimeError("path must be a pathlib.Path")
        ext = path.suffix
        if ext == ".bz2":
            fd = bz2.open(path, self.mode + "t")
        elif ext == ".gz":
            fd = gzip.open(path, self.mode + "t")
        else:
            fd = open(path, self.mode)
        return fd



#!/usr/bin/env python3
"""From a gist by Jannis Mainczyk
https://gist.github.com/jannismain/e96666ca4f059c3e5bc28abb711b5c92
"""
from __future__ import annotations
import json


class CompactJSONEncoder(json.JSONEncoder):
    """A JSON Encoder that puts small containers on single lines."""

    CONTAINER_TYPES = (list, tuple, dict)
    """Container datatypes include primitives or other containers."""

    MAX_WIDTH = 80
    """Maximum width of a container that might be put on a single line."""

    MAX_ITEMS = 6
    """Maximum number of items in container that might be put on single line."""

    def __init__(self, *args, **kwargs):
        # using this class without indentation is pointless
        if kwargs.get("indent") is None:
            kwargs["indent"] = 4
        super().__init__(*args, **kwargs)
        self.indentation_level = 0

    def encode(self, o):
        """Encode JSON object *o* with respect to single line lists."""
        if isinstance(o, (list, tuple)):
            return self._encode_list(o)
        if isinstance(o, dict):
            return self._encode_object(o)
        if isinstance(o, float):  # Use scientific notation for floats
            return format(o, ".12g")
        return json.dumps(
            o,
            skipkeys=self.skipkeys,
            ensure_ascii=self.ensure_ascii,
            check_circular=self.check_circular,
            allow_nan=self.allow_nan,
            sort_keys=self.sort_keys,
            indent=self.indent,
            separators=(self.item_separator, self.key_separator),
            default=self.default if hasattr(self, "default") else None,
        )

    def _encode_list(self, o):
        if self._put_on_single_line(o):
            return "[" + ", ".join(self.encode(el) for el in o) + "]"
        self.indentation_level += 1
        output = [self.indent_str + self.encode(el) for el in o]
        self.indentation_level -= 1
        return "[\n" + ",\n".join(output) + "\n" + self.indent_str + "]"

    def _encode_object(self, o):
        if not o:
            return "{}"
        if self._put_on_single_line(o):
            return (
                "{ "
                + ", ".join(
                    f"{self.encode(k)}: {self.encode(el)}" for k, el in o.items()
                )
                + " }"
            )
        self.indentation_level += 1
        output = [
            f"{self.indent_str}{json.dumps(k)}: {self.encode(v)}" for k, v in o.items()
        ]

        self.indentation_level -= 1
        return "{\n" + ",\n".join(output) + "\n" + self.indent_str + "}"

    def iterencode(self, o, **kwargs):
        """Required to also work with `json.dump`."""
        return self.encode(o)

    def _put_on_single_line(self, o):
        return (
            self._primitives_only(o)
            and len(o) <= self.MAX_ITEMS
            and len(str(o)) - 2 <= self.MAX_WIDTH
        )

    def _primitives_only(self, o: list | tuple | dict):
        if isinstance(o, (list, tuple)):
            return not any(isinstance(el, self.CONTAINER_TYPES) for el in o)
        elif isinstance(o, dict):
            return not any(isinstance(el, self.CONTAINER_TYPES) for el in o.values())

    @property
    def indent_str(self) -> str:
        if isinstance(self.indent, int):
            return " " * (self.indentation_level * self.indent)
        elif isinstance(self.indent, str):
            return self.indentation_level * self.indent
        else:
            raise ValueError(
                f"indent must either be of type int or str (is: {type(self.indent)})"
            )


if __name__ == "__main__":
    import sys

    if "--example" in sys.argv:
        data = {
            "compact_object": {"first": "element", "second": 2},
            "compact_list": ["first", "second"],
            "long_list": ["this", "is", "a", "rather", "long\nlist"],
            "non_ascii": "汉语",
        }
        json.dump(data, sys.stdout, cls=CompactJSONEncoder, ensure_ascii=False)
        exit()

    json.dump(json.load(sys.stdin), sys.stdout)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `seamm_util` package, units module."""

from seamm_util import Q_


def test_substance():
    """Testing that we can convert amounts of substance."""
    E = Q_(1.0, "mol")
    E2 = E.to("")
    assert f"{E:~} = {E2:~.4}" == "1.0 mol = 6.022e+23"


def test_inverse_substance():
    """Testing that we can convert amounts of substance."""
    E = Q_(6.022e23, "")
    E2 = E.to("mol")
    assert f"{E:~} = {E2:~.4}" == "6.022e+23 = 1.0 mol"


def test_energy():
    """Testing that we can convert amounts of substance."""
    E = Q_(1.0, "kcal/mol")
    E2 = E.to("eV")
    assert f"{E:~} = {E2:~.4}" == "1.0 kcal / mol = 0.04336 eV"


def test_inverse_energy():
    """Testing that we can convert amounts of substance."""
    E = Q_(1.0, "eV")
    E2 = E.to("kcal/mol")
    assert f"{E:~} = {E2:~.4}" == "1.0 eV = 23.06 kcal / mol"


def test_force():
    """Testing that we can convert amounts of substance."""
    E = Q_(1, "kcal/mol/angstrom")
    E2 = E.to("eV/angstrom")
    assert f"{E:~} = {E2:~.4}" == "1 kcal / Å / mol = 0.04336 eV / Å"


def test_inverse_force():
    """Testing that we can convert amounts of substance."""
    E = Q_(1, "eV/angstrom")
    E2 = E.to("kcal/mol/angstrom")
    assert f"{E:~} = {E2:~.4}" == "1 eV / Å = 23.06 kcal / Å / mol"



# -*- coding: utf-8 -*-

"""
Provides a Dictionary, identical to a dict, but which can be subclassed safely.
"""

import collections
import collections.abc
import copy
import sys


class Dictionary(collections.abc.MutableMapping):
    """Class that acts like a dictionary, but can safely be subclassed.

    The Dictionary class implements a dictionary-like class that other
    classes can inherit from. In addition, __init__ accepts a keyword
    'ordered' specifying the use of an ordered dictionary.
    """

    @classmethod
    def fromkeys(cls, iterable, value=None):
        """Create a new Dictionary with keys from iterable and values set to
        value.
        """
        result = cls()
        result.data = dict.fromkeys(iterable, value)
        return result

    def __init__(self, *args, ordered=False, **kwargs):
        """Initialize a Dictionary, optionally with data.

        Parameters
        ----------
        args : dict
            0+ dictionaries to update from, in order

        ordered : bool, optional
            A Boolean indicating whether to use an ordered dictionary.

        kwargs : keyword - value pairs
            0+ named arguments to update from
        """
        if ordered:
            # Before python 3.7 dictionaries were not ordered.
            if sys.version_info >= (3, 7):
                self.data = dict()
            else:
                self.data = collections.OrderedDict()
        else:
            self.data = dict()

        self.update(*args, **kwargs)

    # Provide dict like access to the data

    def __getitem__(self, key):
        """Allow [] access to the data."""
        return self.data[key]

    def __setitem__(self, key, value):
        """Allow x[key] access to the data."""
        self.data[key] = value

    def __delitem__(self, key):
        """Allow deletion of keys."""
        del self.data[key]

    def __iter__(self):
        """Allow iteration over the object."""
        return iter(self.data)

    def __len__(self):
        """The len() command."""
        return len(self.data)

    def __str__(self):
        """Printable string representation."""
        return str(self.data)

    def __repr__(self):
        """Official string representation."""
        return repr(self.data)

    def copy(self):
        """Return a shallow copy."""
        return copy.copy(self)



# -*- coding: utf-8 -*-

"""Localize the unit handling."""

import pint

# Default units to use for a dimensionality
_default_units = {
    "[current] * [time]": [
        "e",
    ],
    "[current] * [length] * [time]": [
        "D",
        "e*Å",
        "e*a_0",
        "e*nm",
    ],
    "dimensionless": [
        "degree",
    ],
    "[length]": [
        "Å",
        "pm",
        "nm",
        "um",
        "a_0",
    ],
    "[length] ** 2": [
        "Å^2",
        "pm^2",
        "nm^2",
        "um^2",
        "a_0^2",
    ],
    "[length] ** 3": [
        "Å^3",
        "pm^3",
        "nm^3",
        "um^3",
        "a_0^2",
    ],
    "1 / [length]": [
        "1/Å",
        "1/pm",
        "1/nm",
        "1/um",
        "1/a_0",
    ],
    "[length] ** 2 / [time]": [
        "cSt",
        "St",
        "m^2/s",
        "Å^2/fs",
    ],
    "[length] ** 2 * [mass]": [
        "10^-40*g*cm^2",
        "Da*Å^2",
        "Da*a_0^2",
    ],
    "[length] ** 2 * [mass] / [substance] / [time] ** 2": [
        "kJ/mol",
        "kcal/mol",
        "eV",
        "E_h",
        "Ry",
    ],
    "[length] ** 2 * [mass] / [time] ** 2": [
        "kJ/mol",
        "kcal/mol",
        "eV",
        "E_h",
        "Ry",
    ],
    "[length] * [mass] / [substance] / [time] ** 2": [
        "kcal/mol/Å",
        "kJ/mol/Å",
        "eV/Å",
        "E_h/Å",
        "Ry/Å",
        "E_h/a_0",
        "Ry/a_0",
    ],
    "[mass] / [substance] / [time] ** 2": [
        "kcal/mol/Å^2",
        "kJ/mol/Å^2",
        "eV/Å^2",
        "E_h/Å^2",
        "Ry/Å^2",
        "E_h/a_0^2",
        "Ry/a_0^2",
    ],
    "[length] * [mass] / [time] ** 2": [
        "kcal/mol/Å",
        "kJ/mol/Å",
        "eV/Å",
    ],
    "[length] / [time]": [
        "km/s",
        "m/s",
        "mph",
        "ft/s",
        "Å/fs",
    ],
    "[length] / [time] ** 2": [
        "m/s^2",
        "ft/s^2",
        "Å/fs^2",
    ],
    "[length] ** 2 * [mass] / [substance] / [temperature] / [time] ** 2": [
        "J/K/mol",
        "cal/K/mol",
    ],
    "[mass]": [
        "Da",
        "amu",
        "g",
        "kg",
        "tonne",
        "lb",
        "ton",
    ],
    "[mass] / [length] ** 3": [
        "g/mL",
        "kg/L",
        "kg/m^3",
        "g/mol/Å^3",
        "g/mol/A_0^3",
    ],
    "[mass] / [length] / [time]": [
        "cP",
        "P",
        "mPa*s",
        "Pa*s",
    ],
    "[mass] / [length] / [time] ** 2": [
        "TPa",
        "GPa",
        "MPa",
        "Pa",
        "atm",
        "bar",
        "psi",
        "ksi",
    ],
    "[length] * [time] ** 2 / [mass]": [
        "1/TPa",
        "1/GPa",
        "1/MPa",
        "1/Pa",
        "1/atm",
        "1/bar",
        "1/psi",
        "1/ksi",
    ],
    "[mass] / [time] ** 2": [
        "mdyne/Å",
        "N/m",
    ],
    "1 / [time] ** 2": [
        "mdyne/Å/Da",
        "N/m/Da",
    ],
    "[substance]": [
        "mol",
    ],
    "[temperature]": [
        "K",
        "°C",
        "°F",
        "°R",
    ],
    "1 / [temperature]": [
        "1/K",
        "1/°C",
        "1/°F",
        "1/°R",
    ],
    "[time]": [
        "fs",
        "ps",
        "ns",
        "us",
        "ms",
        "s",
    ],
}


def sort_dimensions(dimensions):
    """Sort the dimensions.

    Parameters
    ----------
    dimensions : str
        The dimensions.

    Returns
    -------
    str
        The sorted dimensions.
    """
    tmp = dimensions.split(" / ")
    numerator = " * ".join(sorted(tmp[0].split(" * ")))
    if len(tmp) > 1:
        denominator = " / ".join(sorted(tmp[1:]))
        return f"{numerator} / {denominator}"
    return numerator


_default_units = {sort_dimensions(k): v for k, v in _default_units.items()}

# Unit handling!
ureg = pint.UnitRegistry(auto_reduce_dimensions=True)
ureg.formatter.default_format = "~P"
pint.set_application_registry(ureg)
Q_ = ureg.Quantity
units_class = ureg("1 km").__class__

if True:
    _d = pint.Context("chemistry")

    factor = ureg.mol / ureg.avogadro_number

    _d.add_transformation("", "[substance]", lambda ureg, x: x * factor)
    _d.add_transformation("[substance]", "", lambda ureg, x: x / factor)

    _d.add_transformation("1 / [substance]", "", lambda ureg, x: x * factor)
    _d.add_transformation("", "1 / [substance]", lambda ureg, x: x / factor)

    _d.add_transformation("[mass] / [substance]", "[mass]", lambda ureg, x: x * factor)
    _d.add_transformation("[mass]", "[mass] / [substance]", lambda ureg, x: x / factor)

    # kJ/mol/Å --> eV/Å
    _d.add_transformation(
        "[length] * [mass] / [substance] / [time] ** 2",
        "[length] * [mass] / [time] ** 2",
        lambda ureg, x: x * factor,
    )
    _d.add_transformation(
        "[length] * [mass] / [time] ** 2",
        "[length] * [mass] / [substance] / [time] ** 2",
        lambda ureg, x: x / factor,
    )

    # kJ/mol --> eV
    _d.add_transformation(
        "[length] ** 2 * [mass] / [substance] / [time] ** 2",
        "[length] ** 2 * [mass] / [time] ** 2",
        lambda ureg, x: x * factor,
    )
    _d.add_transformation(
        "[length] ** 2 * [mass] / [time] ** 2",
        "[length] ** 2 * [mass] / [substance] / [time] ** 2",
        lambda ureg, x: x / factor,
    )

    # kJ/mol/K --> eV/K
    _d.add_transformation(
        "[length] ** 2 * [mass] / [substance] / [temperature] / [time] ** 2",
        "[length] ** 2 * [mass] / [temperature] / [time] ** 2",
        lambda ureg, x: x * factor,
    )
    _d.add_transformation(
        "[length] ** 2 * [mass] / [temperature] / [time] ** 2",
        "[length] ** 2 * [mass] / [substance] / [temperature] / [time] ** 2",
        lambda ureg, x: x / factor,
    )

    # kJ/mol/Å^2 --> eV/Å^2
    _d.add_transformation(
        "[mass] / [substance] / [time] ** 2",
        "[mass] / [time] ** 2",
        lambda ureg, x: x * factor,
    )

    # eV/Å^2 --> kJ/mol/Å^2
    _d.add_transformation(
        "[mass] / [time] ** 2",
        "[mass] / [substance] / [time] ** 2",
        lambda ureg, x: x / factor,
    )

    # kJ/mol/Å^3 --> eV/Å^3
    _d.add_transformation(
        "[mass] / [length] / [substance] / [time] ** 2",
        "[mass] / [length] / [time] ** 2",
        lambda ureg, x: x * factor,
    )

    # eV/Å^3 --> kJ/mol/Å^3
    _d.add_transformation(
        "[mass] / [length] / [time] ** 2",
        "[mass] / [length] / [substance] / [time] ** 2",
        lambda ureg, x: x / factor,
    )

    # kJ/mol/Å^4 --> eV/Å^4
    _d.add_transformation(
        "[mass] / [length] ** 2 / [substance] / [time] ** 2",
        "[mass] / [length] ** 2 / [time] ** 2",
        lambda ureg, x: x * factor,
    )

    # eV/Å^4 --> kJ/mol/Å^4
    _d.add_transformation(
        "[mass] / [length] ** 2 / [time] ** 2",
        "[mass] / [length] ** 2 / [substance] / [time] ** 2",
        lambda ureg, x: x / factor,
    )

    # kJ/mol*Å^6 --> eV*Å^6
    _d.add_transformation(
        "[length] ** 8 * [mass] / [substance] / [time] ** 2",
        "[length] ** 8 * [mass] / [time] ** 2",
        lambda ureg, x: x * factor,
    )

    # eV*Å^6 --> kJ/mol*Å^6
    _d.add_transformation(
        "[length] ** 8 * [mass] / [time] ** 2",
        "[length] ** 8 * [mass] / [substance] / [time] ** 2",
        lambda ureg, x: x / factor,
    )

    ureg.add_context(_d)

ureg.enable_contexts("spectroscopy", "boltzmann", "energy", "chemistry")


def default_units(units_or_dimensions):
    """Return the default units.

    Parameters
    ----------
    units_or_dimensions : str
        The units or dimensionality.

    Returns
    -------
    [str]
        The list of unit strings.
    """
    if units_or_dimensions == "all":
        result = []
        for values in _default_units.values():
            result.extend(values)
        return result

    if "[" in units_or_dimensions or units_or_dimensions == "dimensionless":
        dimensions = units_or_dimensions
    else:
        dimensions = str(Q_(units_or_dimensions).dimensionality)

    dimensions = sort_dimensions(dimensions)

    if dimensions in _default_units:
        return _default_units[dimensions]
    else:
        result = []
        try:
            for units in ureg.get_compatible_units(dimensions):
                result.append(f"{units:~P}")
            tmp = "\n\t".join(result)
            print(
                f"Automatic defaults for '{units_or_dimensions}' ({dimensions}) "
                f"\n\t{tmp}"
            )
        except Exception:
            print(
                f"Warning: can't handle units '{units_or_dimensions}' --> "
                f"{dimensions} for default units."
            )
        return result


if __name__ == "__main__":  # pragma: no cover
    for key in _default_units:
        if key != sort_dimensions(key):
            print(f"{key} -> {sort_dimensions(key)}")
    print()

    E = Q_(1.0, "kcal/mol")
    E2 = E.to("eV")
    print(f"{E:~} = {E2:~.4}")
    print()

    E2 = E.to("kJ/mol")
    print(f"{E:~} = {E2:~.4}")
    print()

    E = Q_(1.0, "eV")
    E2 = E.to("kcal/mol")
    print(f"{E:~} = {E2:~.4}")
    print()

    E2 = E.to("kJ/mol")
    print(f"{E:~} = {E2:~.4}")
    print()

    E = Q_(1.0, "mol")
    E2 = E.to("")
    print(f"{E:~} = {E2:~.4}")
    print()

    E = Q_(6.022e23, "")
    E2 = E.to("mol")
    print(f"{E:~} = {E2:~.4}")
    print()

    E = Q_(1, "kcal/mol/angstrom")
    E2 = E.to("eV/angstrom")
    print(f"{E:~} = {E2:~.4}")
    print()

    E = Q_(1, "eV/angstrom")
    E2 = E.to("kcal/mol/angstrom")
    print(f"{E:~} = {E2:~.4}")
    print()



"""Extend JSON encoding/decoding to handle pint Quantities, as well
as datetimes and timedeltas

To encode:
    import seamm_util
    dump = json.dumps(<variable>, cls=seamm_util.JSONEncoder)

and to decode:
    import seamm_util
    decoder = seamm_util.JSONDecoder()
    <variable> = decoder.decode(dump)

Adapted from
http://taketwoprogramming.blogspot.com/2009/06/subclassing-jsonencoder-and-jsondecode
"""

from seamm_util import ureg, Q_, units_class  # nopep8
import datetime
import json


class JSONEncoder(json.JSONEncoder):
    """
    Encodes a python object, where pint Quantities, datetime and
    timedelta objects are converted into objects that can be decoded
    using the seamm_util.JSONDecoder.

    Adapted from
    http://taketwoprogramming.blogspot.com/2009/06/subclassing-jsonencoder-and-jsondecoder.html  # noqa: E501
    """

    def default(self, obj):
        import seamm

        if isinstance(obj, units_class):
            return {"__type__": "pint_units", "data": obj.to_tuple()}
        elif isinstance(obj, datetime.datetime):
            return {
                "__type__": "datetime",
                "year": obj.year,
                "month": obj.month,
                "day": obj.day,
                "hour": obj.hour,
                "minute": obj.minute,
                "second": obj.second,
                "microsecond": obj.microsecond,
            }
        elif isinstance(obj, datetime.timedelta):
            return {
                "__type__": "timedelta",
                "days": obj.days,
                "seconds": obj.seconds,
                "microseconds": obj.microseconds,
            }
        elif isinstance(obj, seamm.Parameter) or isinstance(obj, seamm.Parameters):
            #  Populate the dictionary with object meta data
            obj_dict = {
                "__class__": obj.__class__.__name__,
                "__module__": obj.__module__,
            }

            #  Populate the dictionary with object properties
            obj_dict.update(obj.to_dict())

            return obj_dict
        else:
            return json.JSONEncoder.default(self, obj)


class JSONDecoder(json.JSONDecoder):
    """Decodes a json string, where pint Quantities, datetime and
    timedelta objects were converted into objects using the
    seamm_util.JSONEncoder, back into a python object.
    """

    def __init__(self):
        super().__init__(object_hook=self.dict_to_object)

    def dict_to_object(self, d):
        if "__type__" in d:
            type = d.pop("__type__")
            if type == "pint_units":
                return Q_.from_tuple(d["data"])
            elif type == "datetime":
                return datetime.datetime(**d)
            elif type == "timedelta":
                return datetime.timedelta(**d)
            else:
                # Oops... better put this back together.
                d["__type__"] = type

        if "__class__" in d:
            class_name = d.pop("__class__")

            if class_name in "Parameter":
                # Get the module name from the dict and import it
                module_name = d.pop("__module__")
                module = __import__(module_name)

                # Get the class from the module
                class_ = getattr(module, class_name)

                # Use dictionary unpacking to initialize the object
                return class_(d)
            elif "Parameters" in class_name:
                # Get the module name from the dict and import it
                module_name = d.pop("__module__")
                module = __import__(module_name)

                # Get the class from the module
                class_ = getattr(module, class_name)

                # Use dictionary unpacking to initialize the object
                return class_(data=d)
            else:
                d["__init__class__"] = class_name

        return d


if __name__ == "__main__":
    acel = ureg("9.8 m/s**2")
    print(acel)
    print(acel.__class__)
    print(units_class)

    t = datetime.datetime.now()
    print(t)
    print(t.__class__)

    dct = {"acel": acel, "time": t}

    dump = json.dumps(dct, cls=JSONEncoder)
    print()
    print(dump)
    print()

    decoder = JSONDecoder()
    t2 = decoder.decode(dump)
    print(t2)
    print()
    print(t2["acel"])
    print(t2["acel"].__class__)
    print(t2["time"])
    print(t2["time"].__class__)



# -*- coding: utf-8 -*-

"""Interact with Zenodo via its REST API"""

import collections.abc
import copy
import configparser
import logging
from pathlib import Path
import pprint
import requests

logger = logging.getLogger(__name__)
# logger.setLevel("DEBUG")

upload_types = {
    "publication": "Publication",
    "poster": "Poster",
    "presentation": "Presentation",
    "dataset": "Dataset",
    "image": "Image",
    "video": "Video/Audio",
    "software": "Software",
    "lesson": "Lesson",
    "physicalobject": "Physical object",
    "other": "Other",
}


class Record(collections.abc.Mapping):
    """A class for handling uploading a record to Zenodo.

    Attributes
    ----------
    data : dict()
        The record data from Zenodo. See https://developers.zenodo.org/#depositions
    token : str
        The Zenodo access token for the user.
    metadata : dict()
        The metadata for updating the record.
    """

    def __init__(self, data, token):
        self.data = data
        self.token = token
        self.metadata = {}

    # Provide dict like access to the widgets to make
    # the code cleaner

    def __getitem__(self, key):
        """Allow [] access to the widgets!"""
        return self.data[key]

    def __iter__(self):
        """Allow iteration over the object"""
        return iter(self.data)

    def __len__(self):
        """The len() command"""
        return len(self.data)

    def __str__(self):
        return pprint.pformat(self.data)

    @property
    def authors(self):
        """Synonym for creators"""
        return self.creators

    @authors.setter
    def authors(self, value):
        self.creators = value

    @property
    def conceptdoi(self):
        """The generic concept DOI."""
        if "conceptdoi" in self.data:
            return self.data["conceptdoi"]
        else:
            return None

    @property
    def creators(self):
        """The creators for the record."""
        if "creators" not in self.metadata:
            if "creators" in self.data["metadata"]:
                self.metadata["creators"] = copy.deepcopy(
                    self.data["metadata"]["creators"]
                )
            else:
                self.metadata["creators"] = []
        return self.metadata["creators"]

    @creators.setter
    def creators(self, value):
        self.metadata["creators"] = copy.deepcopy(value)

    @property
    def description(self):
        """The description for the record."""
        if "description" not in self.metadata:
            if "description" in self.data["metadata"]:
                self.metadata["description"] = self.data["metadata"]["description"]
            else:
                return None
        return self.metadata["description"]

    @description.setter
    def description(self, value):
        self.metadata["description"] = value

    @property
    def doi(self):
        """The (prereserved) DOI."""
        if "doi" in self.data and self.data["doi"] != "":
            return self.data["doi"]
        else:
            return self.data["metadata"]["prereserve_doi"]["doi"]

    @property
    def in_progress(self):
        """Whether the deposition is still in progress, i.e. editable.

        Returns
        -------
        bool
        """
        return self.data["state"] == "inprogress"

    @property
    def keywords(self):
        """The keywords for the record."""
        if "keywords" not in self.metadata:
            if "keywords" in self.data["metadata"]:
                self.metadata["keywords"] = copy.deepcopy(
                    self.data["metadata"]["keywords"]
                )
            else:
                self.metadata["keywords"] = []
        return self.metadata["keywords"]

    @keywords.setter
    def keywords(self, value):
        self.metadata["keywords"] = copy.deepcopy(value)

    @property
    def submitted(self):
        """Whether the record has been submitted.

        If so the files can't be changed, but it may be possible to edit the metadata.

        Returns
        -------
        bool
        """
        return self.data["submitted"]

    @property
    def title(self):
        """The title for the record."""
        if "title" not in self.metadata:
            if "title" in self.data["metadata"]:
                self.metadata["title"] = self.data["metadata"]["title"]
            else:
                return None
        return self.metadata["title"]

    @title.setter
    def title(self, value):
        self.metadata["title"] = value

    @property
    def upload_type(self):
        """The type of record in Zenodo."""
        if "upload_type" not in self.metadata:
            if "upload_type" in self.data["metadata"]:
                self.metadata["upload_type"] = self.data["metadata"]["upload_type"]
            else:
                return None
        return self.metadata["upload_type"]

    @upload_type.setter
    def upload_type(self, value):
        if value not in upload_types:
            raise ValueError(
                f"upload_type '{value}' must be one of "
                f"{', '.join(upload_types.keys())}"
            )
        self.metadata["upload_type"] = value

    def add_creator(self, name, affiliation=None, orcid=None, ignore_duplicates=False):
        """Add a creator (author) to the record.

        Parameters
        ----------
        name : str
            The creators name as "family name, other names"
        affiliation : str, optional
            The creators affiliation (University, company,...)
        orcid : str, optional
            The ORCID id of the creator.
        ignore_duplicates : bool = False
            Silently ignore duplicate records.
        """
        # Already exists?
        for creator in self.creators:
            if "orcid" in creator and orcid is None:
                if creator["orcid"] == orcid:
                    if ignore_duplicates:
                        return
                    raise RuntimeError(f"Duplicate entry for creator: {name}")
            elif creator["name"] == name:
                if ignore_duplicates:
                    return
            raise RuntimeError(f"Duplicate entry for creator: {name}")

        creator = {"name": name}
        if affiliation is not None:
            creator["affiliation"] = affiliation
        if orcid is not None:
            creator["orcid"] = orcid
        self.metadata["creators"].append(creator)

    def add_file(self, path, contents=None, binary=False):
        """Add the given file to the record.

        Parameters
        ----------
        path : str or pathlib.Path
            The path to the file to upload.
        binary : bool = False
            Whether to open as a binary file.
        """
        if self.submitted:
            raise RuntimeError("Files cannot be added to a submitted record.")

        if isinstance(path, str):
            path = Path(path).expanduser()

        url = self.data["links"]["bucket"] + "/" + path.name
        headers = {"Authorization": f"Bearer {self.token}"}
        if contents is None:
            mode = "rb" if binary else "r"
            with open(path, mode) as fd:
                response = requests.put(url, data=fd, headers=headers)
        else:
            response = requests.put(url, data=contents, headers=headers)

        if response.status_code != 201:
            raise RuntimeError(
                f"Error in add_file: code = {response.status_code}"
                f"\n\n{pprint.pformat(response.json())}"
            )

        # Add the new file to the metadata
        self.data["files"].append(response.json())

    def add_keyword(self, keyword):
        """Add a keyword to the record.

        Parameters
        ----------
        keyword : str
            The keyword
        """
        # Already exists?
        if keyword not in self.keywords:
            self.metadata["keywords"].append(keyword)

    def download_file(self, filename, path):
        """Download a file to a local copy.

        Parameters
        ----------
        filename : str
            The name of the file.
        path : pathlib.Path
            The path to download the file to. Can be a directory in which case
            the filename is used in that directory.

        Returns
        -------
        pathlib.Path
            The path to the downloaded file.
        """
        if "files" not in self.data:
            raise RuntimeError("There are no files in the record.")

        if isinstance(path, str):
            path = Path(path)

        if path.is_dir():
            out_path = path / filename
        else:
            out_path = path

        headers = {
            "Content-Type": "application/json",
        }
        if self.token is not None:
            headers["Authorization"] = f"Bearer {self.token}"

        for data in self.data["files"]:
            if data["filename"] == filename:
                url = data["links"]["download"]
                response = requests.get(url, headers=headers, stream=True)

                if response.status_code != 200:
                    raise RuntimeError(
                        f"Error in download_file: code = {response.status_code}"
                        f"\n\n{pprint.pformat(response.json())}"
                    )

                with open(out_path, "wb") as fd:
                    for chunk in response.iter_content(chunk_size=128):
                        fd.write(chunk)

                return out_path

        raise RuntimeError(f"File '{filename}' is not part of the deposit.")

    def files(self):
        """List of the files deposited.

        Returns
        -------
        [str]
        """
        if "files" in self.data:
            return [x["filename"] for x in self.data["files"]]
        else:
            return []

    def get_file(self, filename):
        """Get the contents of a file.

        Parameters
        ----------
        filename : str
            The name of the file.

        Returns
        -------
        str or byte
        """
        if "files" not in self.data:
            raise RuntimeError("There are no files in the record.")

        headers = {
            "Content-Type": "application/json",
        }
        if self.token is not None:
            headers["Authorization"] = f"Bearer {self.token}"

        for data in self.data["files"]:
            if data["key"] == filename:
                url = data["links"]["self"]
                response = requests.get(url, headers=headers)

                if response.status_code != 200:
                    raise RuntimeError(
                        f"Error in get_file: code = {response.status_code}"
                        f"\n\n{pprint.pformat(response.json())}"
                    )
                return response.text

        raise RuntimeError(f"File '{filename}' is not part of the deposit.")

    def publish(self):
        """Publish the record on Zenodo.

        This registers the DOI, and after this the files cannot be changed.
        Any new metadata is uploaded before publishing.
        """
        if len(self.metadata) > 0:
            self.update_metadata()

        url = self.data["links"]["publish"]
        headers = {"Authorization": f"Bearer {self.token}"}

        response = requests.post(url, headers=headers)

        if response.status_code != 202:
            raise RuntimeError(
                f"Error in publish_metadata: code = {response.status_code}"
                f"\n\n{pprint.pformat(response.json())}"
            )

        self.data = response.json()

    def remove_file(self, filename):
        """Remove a file.

        Parameters
        ----------
        filename : str
            The name of the file.
        """
        if self.submitted:
            raise RuntimeError("Files cannot be removed from a submitted record.")

        if "files" not in self.data:
            raise RuntimeError("There are no files in the record.")

        headers = {
            "Authorization": f"Bearer {self.token}",
            "Content-Type": "application/json",
        }

        for index, data in enumerate(self.data["files"]):
            if data["filename"] == filename:
                url = data["links"]["self"]
                response = requests.delete(url, headers=headers)

                if response.status_code != 204:
                    raise RuntimeError(
                        f"Error in remove_file: code = {response.status_code}"
                        f"\n\n{pprint.pformat(response.json())}"
                    )

                # Remove the entry from the metadata
                del self.data["files"][index]

                return

        raise RuntimeError(f"File '{filename}' is not part of the deposit.")

    def remove_keyword(self, keyword):
        """Remove a keyword from the record.

        Parameters
        ----------
        keyword : str
            The keyword
        """
        # Doesn't exist?
        if keyword not in self.keywords:
            self.metadata["keywords"].append(keyword)

    def update_metadata(self):
        """Update the metadata for the record in Zenodo."""
        url = self.data["links"]["self"]
        headers = {
            "Authorization": f"Bearer {self.token}",
            "Content-Type": "application/json",
        }

        data = {"metadata": self.metadata}

        response = requests.put(url, json=data, headers=headers)

        if response.status_code != 200:
            raise RuntimeError(
                f"Error in update_metadata: code = {response.status_code}"
                f"\n\n{pprint.pformat(response.json())}"
            )

        self.data = response.json()
        self.metadata = {}


class Zenodo(object):
    def __init__(self, token=None, configfile="~/.seamm.d/seammrc", use_sandbox=False):
        if use_sandbox:
            self.base_url = "https://sandbox.zenodo.org/api"
        else:
            self.base_url = "https://zenodo.org/api"

        self.configfile = configfile
        self.use_sandbox = use_sandbox

    @property
    def token(self):
        """The appropriate token for Zenodo."""
        path = Path(self.configfile).expanduser()
        if not path.exists:
            raise RuntimeError(
                f"You need a {self.configurationfile} file to publish to Zenodo. "
                "See the documentation for more details."
            )

        config = configparser.ConfigParser()
        config.read(path)

        if self.use_sandbox:
            if "SANDBOX" not in config:
                raise RuntimeError(
                    f"There is no [SANDBOX] section in {self.configfile}."
                )
            if "token" not in config["SANDBOX"]:
                raise RuntimeError(
                    "There is no 'token' in the [SANDBOX] section of "
                    f"{self.configfile}."
                )
            token = config["SANDBOX"]["token"]
        else:
            if "ZENODO" not in config:
                raise RuntimeError(
                    f"There is no [ZENODO] section in {self.configfile}."
                )
            if "token" not in config["ZENODO"]:
                raise RuntimeError(
                    "There is no 'token' in the [ZENODO] section of "
                    f"{self.configfile}."
                )
            token = config["ZENODO"]["token"]

        return token

    def add_version(self, _id):
        """Create a new record object for uploading a new version to Zenodo."""
        headers = {
            "Authorization": f"Bearer {self.token}",
            "Content-Type": "application/json",
        }

        url = self.base_url + f"/deposit/depositions/{_id}/actions/newversion"

        logger.debug(f"add_version {url=}")
        logger.debug(headers)

        response = requests.post(url, headers=headers)

        logger.debug(f"{response.status_code=}")
        logger.debug(f"\n{pprint.pformat(response.json())}")

        if response.status_code != 201:
            raise RuntimeError(
                f"Error in add_version: code = {response.status_code}"
                f"\n\n{pprint.pformat(response.json())}"
            )

        result = response.json()

        # The result is for the original DOI, so get the data for the new one
        url = result["links"]["latest_draft"]
        response = requests.get(url, headers=headers)

        if response.status_code != 200:
            raise RuntimeError(
                f"Error in add_version get latest draft: code = {response.status_code}"
                f"\n\n{pprint.pformat(response.json())}"
            )

        result = response.json()

        return Record(result, self.token)

    def create_record(self):
        """Create a new record object for uploading to Zenodo."""
        headers = {
            "Authorization": f"Bearer {self.token}",
            "Content-Type": "application/json",
        }

        url = self.base_url + "/deposit/depositions"
        response = requests.post(url, json={}, headers=headers)

        if response.status_code != 201:
            raise RuntimeError(
                f"Error in create_record: code = {response.status_code}"
                f"\n\n{pprint.pformat(response.json())}"
            )

        result = response.json()

        return Record(result, self.token)

    def get_deposit_record(self, _id):
        """Get an existing deposit record object from Zenodo."""
        url = self.base_url + f"/deposit/depositions/{_id}"
        try:
            headers = {
                "Authorization": f"Bearer {self.token}",
                "Content-Type": "application/json",
            }
        except Exception:
            token = None
            response = requests.get(url, json={})
        else:
            token = self.token
            response = requests.get(url, json={}, headers=headers)

        if response.status_code != 200:
            raise RuntimeError(
                f"Error in get_deposit_record: code = {response.status_code}"
                f"\n\n{pprint.pformat(response.json())}"
            )

        result = response.json()

        return Record(result, token)

    def get_record(self, _id):
        """Get an existing record object from Zenodo."""
        url = self.base_url + f"/api/records/{_id}"
        try:
            headers = {
                "Authorization": f"Bearer {self.token}",
                "Content-Type": "application/json",
            }
        except Exception:
            response = requests.get(url, json={})
        else:
            response = requests.get(url, json={}, headers=headers)

        if response.status_code != 200:
            raise RuntimeError(
                f"Error in get_record: code = {response.status_code}"
                f"\n\n{pprint.pformat(response.json())}"
            )

        result = response.json()

        return Record(result, None)

    def search(
        self,
        authors=None,
        query="",
        communities=None,
        keywords=None,
        title=None,
        description=None,
        all_versions=False,
        size=25,
        page=1,
    ):
        """Search for records in Zenodo."""
        url = self.base_url + "/records/"

        payload = {
            "size": size,
            "page": page,
        }
        if all_versions:
            payload["all_versions"] = 1

        if communities is not None:
            for community in communities:
                query += f' AND +communities:"{community}"'

        if keywords is not None:
            for keyword in keywords:
                query += f' AND +keywords:"{keyword}"'

        payload["q"] = query

        logger.debug("Payload for query request:\n" + pprint.pformat(payload))

        try:
            headers = {
                "Authorization": f"Bearer {self.token}",
            }
        except Exception:
            response = requests.get(url, params=payload)
        else:
            response = requests.get(url, headers=headers, params=payload)

        if response.status_code != 200:
            raise RuntimeError(
                f"Error in search: code = {response.status_code}"
                f"\n\n{pprint.pformat(response.json())}"
            )

        result = response.json()

        records = []
        if "hits" in result:
            hits = result["hits"]
            n_hits = hits["total"]

            logger.debug(f"{n_hits=}")

            for record in hits["hits"]:
                records.append(Record(record, None))

            for record in records:
                logger.debug(f"\t{record['id']}: {record['metadata']['title']}")
        else:
            logger.debug("Query returned no hits!")

        return n_hits, records



# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""A class for reading, updating and writing the configuration file.

This class handles the configuration (.ini) file as text, so that comments
in the file are preserved.
"""

from pathlib import Path


class Configuration(object):
    def __init__(self, path=None):
        self._path = None

        "A dictionary to save the text of the configuration file."
        self._data = {}

        # Set the path, which reads the file if it exsists
        self.path = path

    def __eq__(self, other):
        "Test if this configuration is equal to another in the sense of text."
        return str(self) == str(other)

    def __str__(self):
        """Create the text of the configuration file."""
        return self.to_string()

    @property
    def path(self):
        """The path to the configuration file."""
        return self._path

    @path.setter
    def path(self, value):
        if value is None:
            self._path = None
            self._data = {}
        else:
            new_path = Path(value).expanduser().resolve()
            if new_path != self._path:
                self._path = new_path
                if new_path.exists():
                    self._read()
                else:
                    self._data = {}

    def add_prolog(self, text="", force=False):
        """Add the prolog to the configuration.

        Parameters
        ----------
        text : str = ''
            The body of the prolog, which should be just comments.
        force : bool = True
            Whether to overwrite an existing prolog.
        """
        if "PROLOG" in self._data and not force:
            raise KeyError("The prolog already exists.")
        self._data["PROLOG"] = text.splitlines()

    def add_section(self, name, text="", force=False):
        """Add a new section to the configuration.

        Parameters
        ----------
        name : str
            The name of the section.
        text : str = ''
            The body of the section, which must be properly formatted.
        force : bool = True
            Whether to overwrite an existing section of the same name.
        """
        if self.section_exists(name) and not force:
            raise KeyError(f"Section '{name}' already exists.")
        section = name.lower()
        self._data[section] = [f"[{name}]"]
        lines = text.splitlines()
        if len(lines) > 0:
            tmp = lines[0].strip()
            if tmp[0] == "[" and tmp[-1] == "]":
                if tmp[1:-1] != name:
                    raise ValueError(
                        "Section name doesn't match: '{tmp[1:-1]}' vs '{name}'"
                    )
                lines = lines[1:]
        self._data[section].extend(lines)

    def file_exists(self):
        """Whether the configuration file exists."""
        if self.path is None:
            return False
        else:
            return self.path.exists()

    def from_string(self, text):
        """Replace the contents of the configuration with those from `text`.

        Parameters
        ----------
        text : str
            The configuration data as text.
        """
        self._data = {}
        section = "PROLOG"
        lines = []
        for line in text.splitlines():
            if len(line) > 0:
                # Look for a section name, like [<name>]
                tmp = line.strip()
                if tmp[0] == "[" and tmp[-1] == "]":
                    self._data[section] = lines
                    section = tmp[1:-1].lower()
                    lines = []
            lines.append(line)
        # Put the last section into the data
        self._data[section] = lines

    def get_prolog(self):
        """Return the prolog of the file, if any.

        Returns
        -------
        str
            The prolog of the configuration.
        """
        if "PROLOG" in self._data:
            result = "\n".join(self._data["PROLOG"]) + "\n"
        else:
            result = ""
        return result

    def get_values(self, section):
        """Return the values in a section as a dictionary.

        Returns an empty dictionary if the section does not exist, or if it
        does not contain any keyword definitions. Use `section_exists` to
        differentiate.

        Parameters
        ----------
        section : str
            The name of the section to retrieve.

        Returns
        -------
        dict
            A dictionary of keyword-value pairs, as strings.
        """
        result = {}
        if self.section_exists(section):
            for line in self._data[section.lower()]:
                line = line.strip()
                if len(line) > 0:
                    if line[0] == "#":
                        continue
                    if "=" in line:
                        key, value = line.split("=", maxsplit=1)
                        result[key.strip()] = value.strip()
        return result

    def _read(self):
        """Read the configuration file and split into sections."""
        self.from_string(self.path.read_text())

    def save(self):
        """Save the current configuration to disk."""
        self.path.write_text(str(self))

    def sections(self):
        """Return a list of sections in the configuration.

        Returns
        -------
        [str]
            The list of sections.
        """
        result = sorted(self._data.keys())
        if "PROLOG" in result:
            result.remove("PROLOG")
        if "seamm" in result:
            result.remove("seamm")
            result.insert(0, "seamm")
        if "default" in result:
            result.remove("default")
            result.insert(0, "default")
        return result

    def section_exists(self, section):
        """Return whether a section exits in the configuration.

        Parameters
        ----------
        section : str
            The name of the section.

        Returns
        -------
        bool
            True if the section exists; False otherwise.
        """
        return section is not None and section.lower() in self._data

    def set_value(self, section, key, value, strict=False):
        """Set the key in a section.

        Parameters
        ----------
        section : str
            The section to work with.
        key : str
            The key to set in the section.
        value : str
            The value to set the key to.
        strict : bool = False
            Raise an error is the key does not already exist.
        """
        lines = self._data[section.lower()]
        found = False
        i = 0
        for line in lines:
            if "=" in line:
                if line.strip().split("=", maxsplit=1)[0].strip() == key:
                    found = True
                    if value is None:
                        lines[i] = f"# {line}"
                    else:
                        lines[i] = f"{key} = {value}"
                    break
            i += 1

        if not found:
            if strict:
                raise KeyError(f"'{key}' not in section {section}.")
            else:
                # Maybe it is there but commented...
                found = False
                i = 0
                for line in lines:
                    line = line.lstrip("#").strip()
                    if "=" in line:
                        if line.split("=", maxsplit=1)[0].strip() == key:
                            found = True
                            if value is not None:
                                lines[i] = f"{key} = {value}"
                            break
                    i += 1
                if not found:
                    if value is None:
                        lines.append(f"# {key} = ")
                    else:
                        lines.append(f"{key} = {value}")

    def to_string(self, section=None):
        """Create the text of a section.

        Parameters
        ----------
        section : str
            The name of the section. Defaults to the entire file.
        """
        if section is None:
            result = []
            if "PROLOG" in self._data:
                result.extend(self._data["PROLOG"])
                if result[-1] != "":
                    result.append("")
            for section in self.sections():
                result.extend(self._data[section])
                if result[-1] != "":
                    result.append("")
        else:
            result = self._data[section.lower()]

        return "\n".join(result) + "\n"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `seamm_util` package, plotting module."""

from seamm_util import Figure
import jinja2
import os

# import pytest

# Create the Jinja environment with the path we want
results_path = os.path.join(os.path.dirname(__file__), "results")
template_path = os.path.join(os.path.dirname(__file__), "templates")
env = jinja2.Environment(loader=jinja2.FileSystemLoader(template_path))


def test_simple_plot():
    """Testing that we can create a simple plot."""
    fig = Figure(jinja_env=env, template="line.html_template", title="Simple plot")

    plot = fig.add_plot("Energies")

    x_axis = plot.add_axis("x", label="Time (ps)")
    y_axis = plot.add_axis("y", label="Energy (kcal/mol)", anchor=x_axis)
    x_axis.anchor = y_axis

    plot.add_trace(
        x_axis=x_axis,
        y_axis=y_axis,
        name="Total",
        showlegend="false",
        x0=0,
        dx=1.0,
        xlabel="t",
        xunits="ps",
        y=[110.0, 111.0, 110.0, 109.0, 110.0],
        ylabel="E",
        yunits="kcal/mol",
        color="#20a8d8",
    )

    graph = fig.dumps()

    results_file = os.path.join(results_path, "simple_plot.html")
    try:
        with open(results_file, "r") as fd:
            correct_result = fd.read()
    except Exception:
        correct_result = ""

    if graph != correct_result:
        with open(results_file + "+", "w") as fd:
            fd.write(graph)

    assert graph == correct_result


def test_two_plots():
    """Testing that we can create two plots."""
    fig = Figure(jinja_env=env, template="line.html_template", title="Two plots")

    plot1 = fig.add_plot("plot1")

    x_axis = plot1.add_axis("x", label="Time (ps)")
    y_axis = plot1.add_axis("y", label="Energy (kcal/mol)", anchor=x_axis)
    x_axis.anchor = y_axis

    plot1.add_trace(
        x_axis=x_axis,
        y_axis=y_axis,
        name="Total",
        showlegend="false",
        x0=0,
        dx=1.0,
        xlabel="t",
        xunits="ps",
        y=[110.0, 111.0, 110.0, 109.0, 110.0],
        ylabel="E",
        yunits="kcal/mol",
        color="#20a8d8",
    )

    plot2 = fig.add_plot("plot2")

    x_axis = plot2.add_axis("x", label="Time (ps)")
    y_axis = plot2.add_axis("y", label="Energy (kcal/mol)", anchor=x_axis)
    x_axis.anchor = y_axis

    plot1.add_trace(
        x_axis=x_axis,
        y_axis=y_axis,
        name="Small",
        showlegend="false",
        x0=0,
        dx=1.0,
        xlabel="t",
        xunits="ps",
        y=[10.0, 11.0, 10.0, 9.0, 10.0],
        ylabel="E",
        yunits="kcal/mol",
        color="blue",
    )

    fig.grid_plots("plot1   -     x", "  ^     ^   plot2")  # yapf: disable

    graph = fig.dumps()

    results_file = os.path.join(results_path, "two_plots.html")
    try:
        with open(results_file, "r") as fd:
            correct_result = fd.read()
    except Exception:
        correct_result = ""

    if graph != correct_result:
        with open(results_file + "+", "w") as fd:
            fd.write(graph)

    assert graph == correct_result


def test_writing_plot(tmpdir):
    """Testing that we can create write a plot to disk."""
    fig = Figure(jinja_env=env, template="line.html_template", title="Simple plot")

    plot = fig.add_plot("Energies")

    x_axis = plot.add_axis("x", label="Time (ps)")
    y_axis = plot.add_axis("y", label="Energy (kcal/mol)", anchor=x_axis)
    x_axis.anchor = y_axis

    plot.add_trace(
        x_axis=x_axis,
        y_axis=y_axis,
        name="Total",
        showlegend="false",
        x0=0,
        dx=1.0,
        xlabel="t",
        xunits="ps",
        y=[110.0, 111.0, 110.0, 109.0, 110.0],
        ylabel="E",
        yunits="kcal/mol",
        color="#20a8d8",
    )

    filename = tmpdir / "plot.html"
    fig.dump(filename)

    with open(filename, "r") as fd:
        graph = fd.read()

    results_file = os.path.join(results_path, "simple_plot.html")
    try:
        with open(results_file, "r") as fd:
            correct_result = fd.read()
    except Exception:
        correct_result = ""

    if graph != correct_result:
        with open(results_file + "+", "w") as fd:
            fd.write(graph)

    assert graph == correct_result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `seamm_util` package."""

import seamm_util
from pathlib import Path

datapath = Path(__file__).parent / "data"


def uri_handler(path):
    if str(path).startswith("data:"):
        return (datapath / str(path)[5:]).expanduser().resolve()
    else:
        return Path(path)


def test_open_1file():
    """Testing that we can read a simple file"""
    filepath = datapath / "file1.txt.gz"
    data = ["file1 line 1", "file1 line 2", "file1 line 3", "file1 line 4"]

    i = 0
    with seamm_util.Open(filepath, "r", include="include") as fd:
        for line in fd:
            assert line.strip() == data[i]
            i += 1


def test_open_2files():
    """Testing that the default 'include' works"""
    filepath = datapath / "file_middle.txt.gz"
    data = [
        "file_middle line 1",
        "file_middle line 2",
        "file1 line 1",
        "file1 line 2",
        "file1 line 3",
        "file1 line 4",
        "file_middle line 3",
    ]

    i = 0
    with seamm_util.Open(filepath, "r", include="include") as fd:
        for line in fd:
            assert line.strip() == data[i]
            i += 1


def test_open_uri():
    """Testing that the default 'include' works"""
    filepath = "data:file_uri.txt.gz"
    data = [
        "file_uri line 1",
        "file_uri line 2",
        "file1 line 1",
        "file1 line 2",
        "file1 line 3",
        "file1 line 4",
        "file_uri line 3",
    ]

    i = 0
    with seamm_util.Open(
        filepath, "r", include="include", uri_handler=uri_handler
    ) as fd:
        for line in fd:
            assert line.strip() == data[i]
            i += 1


def test_open_2files_at_end():
    """Testing that the default 'include' works at the end of a file"""
    filepath = datapath / "file_end.txt.gz"
    data = [
        "file_end line 1",
        "file_end line 2",
        "file_end line 3",
        "file1 line 1",
        "file1 line 2",
        "file1 line 3",
        "file1 line 4",
    ]

    i = 0
    with seamm_util.Open(filepath, "r", include="include") as fd:
        for line in fd:
            assert line.strip() == data[i]
            i += 1


def test_lines():
    """Testing the line count for each line, with an included file

    Note that the line count includes the line with 'include'
    """
    filepath = datapath / "file_middle.txt.gz"

    num = [1, 2, 1, 2, 3, 4, 4]
    i = 0
    with seamm_util.Open(filepath, "r", include="include") as fd:
        for line in fd:
            assert fd.lineno == num[i]
            i += 1


def test_total_lines():
    """Testing the total number of lines with an included file

    Note that the line count includes the line with 'include'
    """
    filepath = datapath / "file_middle.txt.gz"

    with seamm_util.Open(filepath, "r", include="include") as fd:
        for line in fd:
            pass
        assert fd.total_lines == 8


def test_nested_includes():
    """Testing an include within an include"""
    filepath = datapath / "file_include1.txt.gz"
    data = [
        "file_include1 line1",
        "file_end line 1",
        "file_end line 2",
        "file_end line 3",
        "file1 line 1",
        "file1 line 2",
        "file1 line 3",
        "file1 line 4",
    ]

    i = 0
    with seamm_util.Open(filepath, "r", include="include") as fd:
        for line in fd:
            assert line.strip() == data[i]
            i += 1


def test_include_empty_file():
    """Testing that including an empty file works"""
    filepath = datapath / "include_empty_file.txt.gz"
    data = ["file_end line 1", "file_end line 2", "file_end line 3"]

    i = 0
    with seamm_util.Open(filepath, "r", include="include") as fd:
        for line in fd:
            assert line.strip() == data[i]
            i += 1


def test_blank_lines():
    """Testing that empty lines and ones with just blanks don't
    cause any problems
    """
    filepath = datapath / "include_blank_lines.txt.gz"
    data = [
        "file_end line 1",
        "",
        "file_end line 3",
        "blank_lines first line",
        "   ",
        "",
        "blank_lines 4th line",
        "   ",
        "",
    ]

    i = 0
    with seamm_util.Open(filepath, "r", include="include") as fd:
        for line in fd:
            assert line[0:-1] == data[i]
            i += 1


def test_stack():
    """Testing the stack of files part way through reading includes"""
    cwd = Path(__file__).parent
    filepath = datapath / "file_include1.txt.gz"
    data = [
        "data/file1.txt.gz:2",
        "data/file_end.txt.gz:4",
        "data/file_include1.txt.gz:2",
    ]

    i = 0
    with seamm_util.Open(filepath, "r", include="include") as fd:
        for line in fd:
            i += 1
            if i == 6:
                stack = []
                for line in fd.stack():
                    stack.append(str(Path(line).relative_to(cwd)).lower())
                print("Stack")
                print("\t" + "\n\t".join(stack))
                print("should be")
                print("\t" + "\n\t".join(data))
                assert stack == data
                break


def test_different_include():
    """Testing that '#include' works"""
    filepath = datapath / "file_#include.txt.gz"
    data = [
        "file_middle line 1",
        "file_middle line 2",
        "file1 line 1",
        "file1 line 2",
        "file1 line 3",
        "file1 line 4",
        "file_middle line 3",
    ]

    i = 0
    with seamm_util.Open(filepath, "r", include="#include") as fd:
        for line in fd:
            assert line.strip() == data[i]
            i += 1


def test_push():
    """Testing that the default 'include' works"""
    filepath = datapath / "file_middle.txt.gz"
    data = [
        "file_middle line 1",
        "file_middle line 2",
        "file1 line 1",
        "file1 line 2",
        "file1 line 3",
        "file1 line 4",
        "file_middle line 3",
        "file1 line 4",
        "file_middle line 3",
    ]

    i = 0
    with seamm_util.Open(filepath, "r", include="include") as fd:
        for line in fd:
            assert line.strip() == data[i]
            i += 1
            if i == 7:
                fd.push(2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `seamm_util` package, plotting module."""

import pytest
from seamm_util import Figure
import sys


def test_create():
    """Testing that we can create a Figure object."""
    fig = Figure()
    assert str(type(fig)) == "<class 'seamm_util.plotting.Figure'>"


def test_set_get_item():
    """Testing setting and then getting an item."""
    fig = Figure()
    fig["a"] = 53
    assert fig["a"] == 53


def test_len():
    """Testing getting the length of a dictionary."""
    fig = Figure()
    fig["a"] = 53
    assert len(fig) == 1


def test_del_item():
    """Testing that we can delete an item."""
    fig = Figure()
    fig["a"] = 53
    del fig["a"]
    assert len(fig) == 0


def test_iter():
    """Testing that we can iterate over a Figure."""
    fig = Figure(ordered=True)
    fig.update({"a": 1, "b": 2, "c": 3})
    keys = []
    for key in fig:
        keys.append(key)

    assert keys == ["a", "b", "c"]


def test_str():
    """Testing that we can get a str representation."""
    fig = Figure(ordered=True)
    fig.update(a=1, b=2, c=3)
    if sys.version_info >= (3, 7):
        assert str(fig) == "{'a': 1, 'b': 2, 'c': 3}"
    else:
        assert str(fig) == "OrderedDict([('a', 1), ('b', 2), ('c', 3)])"


def test_repr():
    """Testing that we can get an official representation."""
    fig = Figure(ordered=True)
    fig.update(a=1, b=2, c=3)
    if sys.version_info >= (3, 7):
        assert repr(fig) == "{'a': 1, 'b': 2, 'c': 3}"
    else:
        assert repr(fig) == "OrderedDict([('a', 1), ('b', 2), ('c', 3)])"


def test_fromkeys():
    """Test the class method fromkeys."""
    fig = Figure.fromkeys(["a", "b", "c"], 1)
    assert repr(fig) == "{'a': 1, 'b': 1, 'c': 1}"


def test_copy():
    """Testing that we can get an official representation."""
    fig = Figure(ordered=True)
    fig.update(a=1, b=2, c=3)

    e = fig.copy()
    fig["b"] = 10

    if sys.version_info >= (3, 7):
        assert str(e) == "{'a': 1, 'b': 10, 'c': 3}"
    else:
        assert str(e) == "OrderedDict([('a', 1), ('b', 10), ('c', 3)])"


def test_template_attribute():
    """Test that we can get the template as an attribute."""
    fig = Figure(template="My Template")

    assert fig.template == "My Template"


def test_set_template_attribute():
    """Test that we can set the template as an attribute."""
    fig = Figure(template="My Template")
    fig.template = "A different template"

    assert fig.template == "A different template"


def test_get():
    """Test that we can 'get' an existing element."""
    fig = Figure({"a": 1, "b": 2, "c": 3})

    assert fig.get("b") == 2


def test_get_default():
    """Test that we 'get' the default for a missing element."""
    fig = Figure({"a": 1, "b": 2, "c": 3})

    assert fig.get("missing") is None


def test_get_default_value():
    """Test that we 'get' the default for a missing element."""
    fig = Figure({"a": 1, "b": 2, "c": 3})

    assert fig.get("missing", "default") == "default"


def test_add_plot():
    """Test that we can add a plot."""
    fig = Figure({"a": 1, "b": 2, "c": 3})
    plot1 = fig.add_plot("plot1")
    assert str(type(plot1)) == "<class 'seamm_util.plotting.Plot'>"


def test_adding_duplicate_plot():
    """Test that adding a duplicate plot raises an error."""
    fig = Figure({"a": 1, "b": 2, "c": 3})
    plot1 = fig.add_plot("plot1")  # noqa: F841
    with pytest.raises(KeyError, match=r"Plot 'plot1' already exists."):
        plot2 = fig.add_plot("plot1")  # noqa: F841


def test_get_plot():
    """Test that we can get an existing plot."""
    fig = Figure({"a": 1, "b": 2, "c": 3})
    plot1 = fig.add_plot("plot1")  # noqa: F841
    plot1a = fig.get_plot("plot1")
    assert str(type(plot1a)) == "<class 'seamm_util.plotting.Plot'>"


def test_getting_missing_plot():
    """Test that getting a missing plot raises an error."""
    fig = Figure({"a": 1, "b": 2, "c": 3})
    plot1 = fig.add_plot("plot1")  # noqa: F841
    with pytest.raises(KeyError, match=r"Plot 'plot2' does not exist."):
        plot2 = fig.get_plot("plot2")  # noqa: F841


def test_grid_error1():
    """Test the gridding of plots, error with initial '-'."""
    fig = Figure()
    with pytest.raises(RuntimeError, match=r"'-' cannot be the first item.*"):
        fig.grid_plots(
            "plot1 plot2 plot3 plot4",
            "  -     -     -     x  ",
            "  ^     ^     ^   plot6",
        )  # yapf: disable


def test_grid_error2():
    """Test the gridding of plots, error with '-' following '^' or 'x'."""
    msg = r"'-' cannot follow 'x' or '^': '  ^     -     ^   plot6'"

    fig = Figure()
    with pytest.raises(RuntimeError) as e:
        fig.grid_plots(
            "plot1 plot2 plot3 plot4",
            "plot5   -     -     x  ",
            "  ^     -     ^   plot6",
        )  # yapf: disable
    assert msg in str(e)


def test_grid_error3():
    """Test the gridding of plots, error with duplicate plots."""
    msg = r"Plot 'plot6' is already used: '  ^     ^     ^   plot6'"

    fig = Figure()
    with pytest.raises(RuntimeError) as e:
        fig.grid_plots(
            "plot1 plot2 plot3 plot4",
            "plot6   -     -     x  ",
            "  ^     ^     ^   plot6",
        )  # yapf: disable
    assert msg in str(e)


def test_grid_error4():
    """Test the gridding of plots, error with rowspan."""
    msg = r"The cell at 2,2 should be a '^', not 'plot7'"

    fig = Figure()
    fig.add_plot("plot1")
    fig.add_plot("plot2")
    fig.add_plot("plot3")
    fig.add_plot("plot4")
    fig.add_plot("plot5")
    fig.add_plot("plot6")
    fig.add_plot("plot7")
    with pytest.raises(RuntimeError) as e:
        fig.grid_plots(
            "plot1 plot2 plot3 plot4",
            "plot6   -     -     x  ",
            "  ^     ^   plot7   -  ",
        )  # yapf: disable
    assert msg in str(e)



# -*- coding: utf-8 -*-

"""Tools for handling the shorthand definitions of lists in SEAMM."""


def parse_list(input, duplicates=True, sort=False, **kwargs):
    """Parse a SEAMM-style list of values, including ranges.

    Parameters
    ----------
    input : str
        The list definition (see below).
    duplicates : bool = True
        Allows duplicate values in the list by default.
    sort : bool = False
        If False, the list is not sorted. True or "ascending" sorts into ascending order
        while "descending" into reverse order. Only the first letter is tested, so any
        abbreviation works.
    kwargs : {str: any}
        Optional dictionary of values of variables used in the list.

    Returns
    -------
    list : [int or float]
        The expanded list of values

    The input list definition consists of any number of individual values or ranges,
    separated by commas (','). Ranges are similar to those in Python *except* that the
    'stop' value is included in the list if stop = start + n*step. Ranges are defined as
    start:stop[:step]. If step is ommitted, it defaults to 1.

    Examples
    --------
    A simple, commma delimited list is transformed into a python list:

    >>> parse_list("1, 2, 4, 5")
    [1, 2, 4, 5]

    If any of the values is a float, the result is also a float

    >>> parse_list("1, 2.0, 4, 5")
    [1, 2.0, 4, 5]

    This is a simple range:

    >>> parse_list("1:6")
    [1, 2, 3, 4, 5, 6]

    Note that the stop value is include in the result. This is different than a python
    range where the termination is <, not <=. This is more natural for users of SEAMM.
    For example, scanning a dihedral angle from 0 to 180 degrees, most people will
    expect 180 to be in the list.

    Now using a step of 2:

    >>> parse_list("1:6:2")
    [1, 3, 5]

    Note that the stop value is not included in the output in this case.

    The step can be negative as long as stop < start:

    >>> parse_list("6:1:-2")
    [6, 4, 2]

    Now we can put it all together:

    >>> parse_list("1, 2.0, 4, 5, 1:6, 1:6:2")
    [1, 2.0, 4, 5, 1, 2, 3, 4, 5, 6, 1, 3, 5]

    Note that a value can appear multiple times in a simple list. This can be controlled
    by setting duplicates = False:

    >>> parse_list("1, 2.0, 4, 5, 1:6, 1:6:2", duplicates=False)
    [1, 2.0, 4, 5, 3, 6]

    You can also sort the results into ascending or descending order:

    >>> parse_list("1, 2.0, 4, 5, 1:6, 1:6:2", sort="ascending")
    [1, 1, 1, 2.0, 2, 3, 3, 4, 4, 5, 5, 5, 6]

    and if you want to remove the duplicates too:

    >>> parse_list("1, 2.0, 4, 5, 1:6, 1:6:2", duplicates=False, sort="descending")
    [6, 5, 4, 3, 2.0, 1]

    Where do variable fit? Any of the values in the input string can be variables as
    as long as values are are given as keyword arguments:

    >>> parse_list("first:last:incr", first=1, last=6, incr=0.5)
    [1, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0, 5.5, 6.0]

    It is often convenient if the variables are in a dictionary:

    >>> values = {"first": 6, "last": 1, "incr":-0.5}
    >>> parse_list("first:last:incr", **values)
    [6, 5.5, 5.0, 4.5, 4.0, 3.5, 3.0, 2.5, 2.0, 1.5]
    """
    _globals = {
        "__builtins__": None,
    }
    result = []
    for part in input.split(","):
        if ":" in part:
            # A range...
            tmp = part.split(":")
            if len(tmp) == 3:
                start, stop, step = tmp
            elif len(tmp) == 2:
                start, stop = tmp
                step = "1"
            else:
                raise RuntimeError(f"Badly formed range '{part}' in parse_list")

            start = eval(start, _globals, kwargs)
            stop = eval(stop, _globals, kwargs)
            step = eval(step, _globals, kwargs)

            # Sanity checks!
            if step == 0:
                raise ValueError("A step of zero is not allow in ranges!")
            elif step < 0 and start < stop:
                raise ValueError(f"If start < stop, the step must be positive: {step}")
            elif step > 0 and start > stop:
                raise ValueError(f"If stop < start, the step must be negative: {step}")

            # Flip the sign of the test if counting down
            sign = -1 if step < 0 else 1
            value = start

            # Add a very small number for roundoff in floats
            while sign * (stop - value + stop / 1.0e8) >= 0:
                if not duplicates:
                    for tmp in result:
                        if abs(value - tmp) < 1.0e-8:
                            break
                    else:
                        result.append(value)
                else:
                    result.append(value)
                value += step
        else:
            value = eval(part, _globals, kwargs)
            result.append(value)

    if isinstance(sort, str):
        if sort[0] == "d":
            result.sort(reverse=True)
        elif sort[0] == "a":
            result.sort()
        else:
            raise ValueError("'sort' must be 'a...' or 'd...' or a bool")
    elif sort:
        result.sort()

    return result


if __name__ == "__main__":
    import doctest

    result = doctest.testmod(verbose=False)



# -*- coding: utf-8 -*-

"""A configuration and argument parser tailored to SEAMM.

The goal of this is to allow options to be specified both on the
command line and in ini files, with command line having
precedence. This is tailored to SEAMM flowcharts so that options for
each type of step (plug-in) is in its own section on the command line
or in the ini file, labeled by the type of section.
"""

import argparse
import configparser
import logging
import os
from pathlib import Path
import sys

# logging.basicConfig(level="WARNING")
logger = logging.getLogger(__name__)

# Used in parser getters to indicate the default behaviour when a specific
# option is not found it to raise an exception. Created to enable `None' as
# a valid fallback value.
_UNSET = object()


class ArgumentParser(object):
    """Replacement and extension for argparse.ArgumentParser that adds
    support for .ini files plus something like the subcommands in
    argparse but with multiple ones active at the same time.

    This is accomplished by separating sections with
    '<section name>' on the commandline, and using
    '[<section name>] in the .ini files. e.g.

    test.flow lammps-step --log-level debug psi4-step --np 32
    """

    def __init__(
        self,
        *args,
        ini_files=[
            os.path.expanduser("~/.seamm.d/seamm.ini"),
            "seamm.ini",
        ],
        interpolation=configparser.ExtendedInterpolation(),
        **kwargs,
    ):
        """Command-line and .ini parser for SEAMM"""

        self._parsers = {}
        self._ini_files = ini_files
        self._interpolation = interpolation
        if self._interpolation is None:
            self._interpolation = configparser.Interpolation()
        self._options = {}
        self._origins = {}
        self._ini_files_used = []

    def add_parser(
        self,
        section,
        prog=None,
        usage=None,
        description=None,
        epilog=None,
        parents=[],
        formatter_class=argparse.HelpFormatter,
        prefix_chars="-",
        fromfile_prefix_chars=None,
        argument_default=None,
        conflict_handler="error",
        add_help=True,
        allow_abbrev=True,
    ):
        """Capture the information needed to create a subparser.

        Parameters
        ----------
        section : str
            The name of this section of the parsing.
        prog : str = sys.argv[0]
            The name of the program
        usage : str = None
            A usage message (default: auto-generated from arguments)
        description : str = None
            A description of what the program does
        epilog : str = None
            Text following the argument descriptions
        parents : [argparse.ArgumentParser] = []
            Parsers whose arguments should be copied into this one
        formatter_class : argparse.HelpFormatter
            HelpFormatter class for printing help messages
        prefix_chars : str = '-'
            Characters that prefix optional arguments
        fromfile_prefix_chars : str = None
            Characters that prefix files containing additional arguments
        argument_default : str = None
            The default value for all arguments
        conflict_handler : str = 'error'
            String indicating how to handle conflicts
        add_help : bool = True
            Add a -h/-help option
        allow_abbrev : bool = True
            Allow long options to be abbreviated unambiguously
        """
        if section in self._parsers:
            raise ValueError(f"Parser '{section}' already exists")
        data = self._parsers[section] = {"arguments": [], "groups": {}}
        data["init_args"] = {
            "prog": prog,
            "usage": usage,
            "description": description,
            "epilog": epilog,
            "parents": parents,
            "formatter_class": formatter_class,
            "prefix_chars": prefix_chars,
            "fromfile_prefix_chars": fromfile_prefix_chars,
            "argument_default": argument_default,
            "conflict_handler": conflict_handler,
            "add_help": add_help,
            "allow_abbrev": allow_abbrev,
        }

    def exists(self, name):
        return name in self._parsers

    def add_argument(self, section, *args, **kwargs):
        if section not in self._parsers:
            raise KeyError(f"Parser '{section}' does not exist.")
        self._parsers[section]["arguments"].append((args, kwargs))

    def add_argument_group(self, section, title, description=None):
        if section not in self._parsers:
            raise KeyError(f"Parser '{section}' does not exist.")
        self._parsers[section]["groups"][title] = {"description": description}

    def parse_args(self, args=None):
        """Parse the .ini file and command-line arguments.

        There are three distinct steps to the process:

            1. Process the command-line arguments, breaking them into
               sections corresponding to the subparsers.
            2. Read the .ini files, if any.
            3. Create the actual parsers from the saved information,
               replacing or adding defaults for any values found in the
               .ini files.
            4. Parse each section of the command-line with the correct parser.

        Parameters
        ----------
        args : [str] = sys.argv[1:]
            The arguments to parse. Defaults to the command-line arguments.
        """
        if args is None:
            args = sys.argv[1:]

        # Check for help!
        if "-h" in args or "--help" in args:
            self.print_help(args)
            exit()

        # 1. Process the command-line arguments, breaking them into
        #    sections corresponding to the subparsers.

        arg_sections = self.split_args(args)

        # 2. Read the .ini files, if any.

        # Moving from ~/SEAMM/seamm.ini to ~/.seamm.d/seamm.ini
        # Make directory and move if it does not exist
        path = Path("~/.seamm.d/seamm.ini").expanduser()
        tmp = Path("~/SEAMM/seamm.ini").expanduser()
        if not path.parent.exists():
            path.parent.mkdir(parents=True)

        if tmp.exists() and not path.exists():
            print("/ntext:")
            print(tmp.read_text())
            print()
            path.write_text(tmp.read_text())
            tmp.unlink()

        # Now read...

        config = configparser.ConfigParser(interpolation=None)
        self._ini_files_used = config.read(self._ini_files)

        # 3. Create the actual parsers from the saved information,
        #    replacing or adding defaults for any values found in the
        #    .ini files.

        self._options = {}
        self._origins = {}
        defaults = {}
        groups = {}
        for section, data in self._parsers.items():
            # The parser
            parser = argparse.ArgumentParser(**data["init_args"])

            # Any groups
            for group, kwargs in data["groups"].items():
                groups[group] = parser.add_argument_group(group, **kwargs)

            # Need to add or replace default if the variable was in a .ini
            # file. This next code is straight for argparse to make sure we
            # have the variable's name in 'dest'.
            for args, kwargs in data["arguments"]:
                if not args or len(args) == 1 and args[0][0] not in parser.prefix_chars:
                    variable = args[0]
                else:
                    kwargs = parser._get_optional_kwargs(*args, **kwargs)
                    variable = kwargs["dest"]
                    args = kwargs["option_strings"]
                    del kwargs["option_strings"]
                # See if it was in the .ini files
                try:
                    tmp = variable.replace("_", "-")
                    if config.has_section(section):
                        default = config.get(section, tmp)
                    else:
                        default = config.get(configparser.DEFAULTSECT, tmp)
                    kwargs["default"] = default
                    # Track what the default is and where it comes from
                    defaults[variable] = {"origin": "configfile"}
                    # May need to convert the type of the default
                    if "type" in kwargs:
                        default = kwargs["type"](default)
                    kwargs["default"] = default
                except Exception:
                    # Track what the default is and where it comes from
                    defaults[variable] = {"origin": "default"}

                # and finally add the argument to the parser.
                group = kwargs.pop("group", None)
                if group is not None:
                    if group not in groups:
                        raise ValueError(f"argument group '{group}' is not defined")
                    groups[group].add_argument(*args, **kwargs)
                else:
                    parser.add_argument(*args, **kwargs)

                defaults[variable]["value"] = parser.get_default(variable)

            # 4. Parse each section of the command-line with the correct parser
            tmp = parser.parse_args(arg_sections[section])
            self._options[section] = vars(tmp)

            origin = self._origins[section] = {}
            for variable, value in self._options[section].items():
                if variable in defaults:
                    if value == defaults[variable]["value"]:
                        origin[variable] = defaults[variable]["origin"]
                    else:
                        origin[variable] = "commandline"
                else:
                    origin[variable] = "commandline"

        # Handle any interpolations, putting the results in a temporary mirror
        tmp = {}
        for section, options in self._options.items():
            data = tmp[section] = {}
            for option, value in options.items():
                if isinstance(value, str):
                    value = self._interpolation.before_get(
                        self, section, option, value, options
                    )
                data[option] = value

        # Override the options with the mirror.
        self._options = tmp

        return self._options

    def get(self, section, option, *, raw=False, vars=None, fallback=_UNSET):
        """Get an option value for a given section.

        If `vars' is provided, it must be a dictionary. The option is looked up
        in `vars' (if provided), `section', and in `DEFAULTSECT' in that order.
        If the key is not found and `fallback' is provided, it is used as
        a fallback value. `None' can be provided as a `fallback' value.

        If interpolation is enabled and the optional argument `raw' is False,
        all interpolations are expanded in the return values.

        Arguments `raw', `vars', and `fallback' are keyword only.

        The section DEFAULT is special.
        """
        try:
            d = self._options[section]
        except KeyError:
            if fallback is _UNSET:
                raise
            else:
                return fallback
        option = self.optionxform(option)
        try:
            value = d[option]
        except KeyError:
            if fallback is _UNSET:
                raise configparser.NoOptionError(option, section)
            else:
                return fallback

        if raw or value is None:
            return value
        else:
            return self._interpolation.before_get(self, section, option, value, d)

    def get_options(self, section=None):
        """Return the options for the given section.

        Parameters
        ----------
        section : str = None
            Return the options for the given section, or all sections

        Returns
        -------
        {str: int, float, str or bool}
            The dictionary of options for the section, or all sections.
        """
        if section is None:
            return {**self._options}
        else:
            return {**self._options[section]}

    def get_origins(self, section=None):
        """Return the origins of the options for the given section.

        Parameters
        ----------
        section : str = None
            Return the origins for the given section, or all sections

        Returns
        -------
        {str: str}
            The dictionary of origins for the section, or all sections.
        """
        if section is None:
            return {**self._origins}
        else:
            return {**self._origins[section]}

    def get_ini_files(self):
        """Return a list of the .ini files actually used.

        Returns
        -------
        [str] :
            The list of .ini files found
        """
        return [*self._ini_files_used]

    def optionxform(self, optionstr):
        return optionstr.lower()

    def print_help(self, args=None, fd=sys.stdout):
        fd.write(self.format_help(args))

    def format_help(self, args=None):
        """Prepare the help message.

        Parameters
        ----------
        args : [str] = sys.argv[1:]
            The arguments to pass the parsers, defaults to sys.argv[1:]
        """
        if args is None:
            args = sys.argv[1:]

        # 1. Process the command-line arguments, breaking them into
        #    sections corresponding to the subparsers.

        # parser = argparse.ArgumentParser(prefix_chars='+')
        # parser.add_argument('SEAMM', nargs='*', default='')
        # for section in self._parsers:
        #     parser.add_argument(f'++{section}', nargs='*', default='')
        # arg_sections = vars(parser.parse_args(args))
        # del parser

        arg_sections = self.split_args(args)

        # What to do depends on where the --help (or -h) flag is. If it is in
        # the SEAMM section, print its help plus the plug-ins. Otherwise,
        # print the help for that plug-in.

        section_args = arg_sections["SEAMM"]
        if "--help" in section_args or "-h" in section_args:
            # set up any SEAMM options
            groups = {}
            if "SEAMM" in self._parsers:
                data = self._parsers["SEAMM"]

                # The parser
                parser = argparse.ArgumentParser(**data["init_args"])

                # Any groups
                for group, kwargs in data["groups"].items():
                    groups[group] = parser.add_argument_group(group, **kwargs)

                # and the arguments
                for args, kwargs in data["arguments"]:
                    group = kwargs.pop("group", None)
                    if group is not None:
                        if group not in groups:
                            raise ValueError(f"argument group '{group}' is not defined")
                        groups[group].add_argument(*args, **kwargs)
                    else:
                        parser.add_argument(*args, **kwargs)
            else:
                parser = argparse.ArgumentParser()

            # and then the plug-ins
            plugins = parser.add_argument_group(
                "plug-ins",
                description=(
                    "The plug-ins in this flowchart, which have their own " "options."
                ),
            )
            for section, data in self._parsers.items():
                if section == "SEAMM":
                    continue
                init_args = data["init_args"]
                if "description" in init_args:
                    plugins.add_argument(section, help=init_args["description"])
                else:
                    plugins.add_argument(section)
            return parser.format_help()

        # Find the first section what contains '--help', and return its help.
        groups = {}
        for section, data in self._parsers.items():
            section_args = arg_sections[section]
            if "--help" in section_args or "-h" in section_args:
                # The parser
                parser = argparse.ArgumentParser(**data["init_args"])

                # Any groups
                for group, kwargs in data["groups"].items():
                    groups[group] = parser.add_argument_group(group, **kwargs)

                for args, kwargs in data["arguments"]:
                    group = kwargs.pop("group", None)
                    if group is not None:
                        if group not in groups:
                            raise ValueError(f"argument group '{group}' is not defined")
                        groups[group].add_argument(*args, **kwargs)
                    else:
                        parser.add_argument(*args, **kwargs)
                return parser.format_help()

    def items(self, section=_UNSET, raw=False, vars=None):
        """Return a list of (name, value) tuples for each option in a section.

        All % interpolations are expanded in the return values, based on the
        defaults passed into the constructor, unless the optional argument
        `raw' is true.  Additional substitutions may be provided using the
        `vars' argument, which must be a dictionary whose contents overrides
        any pre-existing defaults.

        The section DEFAULT is special.
        """
        if section is _UNSET:
            return super().items()
        try:
            d = self._options[section]
        except KeyError:
            if section != self.default_section:
                raise configparser.NoSectionError(section)
        orig_keys = list(d.keys())
        # Update with the entry specific variables
        if vars:
            for key, value in vars.items():
                d[self.optionxform(key)] = value
        if raw:

            def value_getter(option):
                return d[option]

        else:

            def value_getter(option):
                return self._interpolation.before_get(
                    self, section, option, d[option], d
                )

        return [(option, value_getter(option)) for option in orig_keys]

    def split_args(self, args=None):
        """Break the argument line into sections for each plug-in.

        Parameters
        ----------
        args : [str] = sys.argv[1:]

        Returns
        -------
        dict[str: [str]]
            A dictionary with entries of an array of strings for all plug-ins
            plus 'SEAMM'
        """
        sections = self._parsers.keys()
        result = {x: [] for x in sections}
        result["SEAMM"] = []

        section = "SEAMM"
        for arg in args:
            if arg in sections:
                section = arg
            else:
                result[section].append(arg)

        return result


#
# Keep a list of named parsers and a method to create/return one of them
#
_parsers = {}


def getParser(name="SEAMM", **kwargs):
    """Get the named command-line argument parser.

    Returns
    -------
    ArgumentParser
        The seamm.ArgumentParser for handling commandline and
        config-file parsing.
    """

    if name not in _parsers:
        logger.info(f"Creating parser '{name}'.")
        _parsers[name] = ArgumentParser(**kwargs)

    logger.info(f"Returning parser '{name}' ({_parsers[name]})")
    return _parsers[name]


def seamm_parser(name="SEAMM"):
    """Setup the command-line argument parser for SEAMM.

    Returns
    -------
    ArgumentParser
        The seamm.ArgumentParser for handling commandline and
        config-file parsing.
    """
    parser = getParser(name=name)

    if parser.exists("SEAMM"):
        return parser

    parser.add_parser(
        "SEAMM",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        prog=sys.argv[0],
    )

    # Main options
    parser.add_argument_group(
        "SEAMM",
        "main options",
        "The main options for SEAMM",
    )

    parser.add_argument(
        "SEAMM",
        "--root",
        group="main options",
        dest="root",
        default="~/SEAMM",
        action="store",
        help="The root directory for SEAMM data, default: %(default)s",
    )
    parser.add_argument(
        "SEAMM",
        "--datastore",
        group="main options",
        dest="datastore",
        default="${root}/Jobs",
        action="store",
        help="The datastore (directory) for this run, default: %(default)s",
    )
    parser.add_argument(
        "SEAMM",
        "--job-id-file",
        group="main options",
        dest="job_id_file",
        default=None,
        action="store",
        help="The job_id file to use.",
    )
    parser.add_argument(
        "SEAMM",
        "--dashboards",
        group="main options",
        dest="dashboards",
        default="${root}/dashboards.ini",
        action="store",
        help="The configuration file for accessible dashboards: %(default)s",
    )

    # Debugging options
    parser.add_argument_group(
        "SEAMM",
        "debugging options",
        "Options for turning on debugging output and tools",
    )

    parser.add_argument(
        "SEAMM",
        "--log-level",
        group="debugging options",
        default="WARNING",
        type=str.upper,
        choices=["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="The level of informational output, default: '%(default)s'",
    )

    # add options for the job
    parser.add_argument_group("SEAMM", "job options", "Options for jobs")

    parser.add_argument(
        "SEAMM",
        "--database",
        group="job options",
        default="seamm.db",
        help="The database for this job.",
    )
    parser.add_argument(
        "SEAMM",
        "--read-only",
        group="job options",
        action="store_true",
        help="Whether to open the database as read-only.",
    )
    parser.add_argument(
        "SEAMM",
        "--standalone",
        group="job options",
        action="store_true",
        help="Run this workflow as-is without using the job, etc.",
    )
    parser.add_argument(
        "SEAMM",
        "--project",
        group="job options",
        dest="projects",
        action="append",
        help="The project(s) for this job.",
    )
    parser.add_argument(
        "SEAMM",
        "--title",
        group="job options",
        dest="title",
        default="",
        action="store",
        help="The title for this run.",
    )
    parser.add_argument(
        "SEAMM",
        "--description",
        group="job options",
        dest="description",
        default="",
        action="store",
        help="The longer description for this run.",
    )
    parser.add_argument(
        "SEAMM",
        "--force",
        group="job options",
        dest="force",
        action="store_true",
        help="Overwrite the job output if it exists.",
    )

    # Hardware options
    parser.add_argument_group(
        "SEAMM",
        "hardware options",
        (
            "Options about memory limits, parallelism and other details "
            "connected with hardware."
        ),
    )

    parser.add_argument(
        "SEAMM",
        "--parallelism",
        group="hardware options",
        default="any",
        choices=["none", "mpi", "openmp", "any"],
        help="Whether to limit parallel usage to certain types.",
    )

    parser.add_argument(
        "SEAMM",
        "--ncores",
        group="hardware options",
        default="available",
        help=(
            "The maximum number of cores/threads to use in any step. "
            "Default: all available cores."
        ),
    )

    parser.add_argument(
        "SEAMM",
        "--memory",
        group="hardware options",
        default="available",
        help=(
            "The maximum amount of memory to use in any step, which can be "
            "'all' or 'available', or a number, which may use k, Ki, "
            "M, Mi, etc. suffixes. Default: available."
        ),
    )

    return parser



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `seamm_util` package, Dictionary module."""

from seamm_util import Dictionary
import sys


def test_create():
    """Testing that we can create a Dictionary object."""
    d = Dictionary()
    assert str(type(d)) == "<class 'seamm_util.dictionary.Dictionary'>"


def test_set_get_item():
    """Testing setting and then getting an item."""
    d = Dictionary()
    d["a"] = 53
    assert d["a"] == 53


def test_len():
    """Testing getting the length of a dictionary."""
    d = Dictionary()
    d["a"] = 53
    assert len(d) == 1


def test_del_item():
    """Testing that we can delete an item."""
    d = Dictionary()
    d["a"] = 53
    del d["a"]
    assert len(d) == 0


def test_iter():
    """Testing that we can iterate over a Dictionary."""
    d = Dictionary(ordered=True)
    d.update({"a": 1, "b": 2, "c": 3})
    keys = []
    for key in d:
        keys.append(key)

    assert keys == ["a", "b", "c"]


def test_str():
    """Testing that we can get a str representation."""
    d = Dictionary(ordered=True)
    d.update(a=1, b=2, c=3)
    if sys.version_info >= (3, 7):
        assert str(d) == "{'a': 1, 'b': 2, 'c': 3}"
    else:
        assert str(d) == "OrderedDict([('a', 1), ('b', 2), ('c', 3)])"


def test_repr():
    """Testing that we can get an official representation."""
    d = Dictionary(ordered=True)
    d.update(a=1, b=2, c=3)
    if sys.version_info >= (3, 7):
        assert repr(d) == "{'a': 1, 'b': 2, 'c': 3}"
    else:
        assert repr(d) == "OrderedDict([('a', 1), ('b', 2), ('c', 3)])"


def test_fromkeys():
    """Test the class method fromkeys."""
    d = Dictionary.fromkeys(["a", "b", "c"], 1)
    assert repr(d) == "{'a': 1, 'b': 1, 'c': 1}"


def test_copy():
    """Testing that we can get an official representation."""
    d = Dictionary(ordered=True)
    d.update(a=1, b=2, c=3)

    e = d.copy()
    d["b"] = 10

    if sys.version_info >= (3, 7):
        assert repr(e) == "{'a': 1, 'b': 10, 'c': 3}"
    else:
        assert repr(e) == "OrderedDict([('a', 1), ('b', 10), ('c', 3)])"



# -*- coding: utf-8 -*-

"""seamm_util
Utility functions for the SEAMM environment.
"""

# Bring up the classes so that they appear to be directly in
# the seamm_util package.

from .argument_parser import getParser  # noqa: F401
from .argument_parser import seamm_parser  # noqa: F401
from .compact_json_encoder import CompactJSONEncoder  # noqa: F401
from .configuration import Configuration  # noqa: F401
from .elemental_data import element_data  # noqa: F401
from .check_executable import check_executable  # noqa: F401
from .dictionary import Dictionary  # noqa: F401
from .list_definition import parse_list  # noqa: F401
from .units import ureg, Q_, units_class, default_units  # noqa: F401
from .include_open import Open  # noqa: F401
from .include_open import splitext  # noqa: F401
from .seamm_json import JSONDecoder  # noqa: F401
from .seamm_json import JSONEncoder  # noqa: F401
from .plotting import Figure  # noqa: F401
from . import printing  # noqa: F401
from . import variable_names  # noqa: F401
from . import water_models  # noqa: F401
from . import zenodo  # noqa: F401
from .zenodo import Zenodo  # noqa: F401

# Handle versioneer
from ._version import get_versions

__author__ = """Paul Saxe"""
__email__ = "psaxe@molssi.org"
versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions
